import pytest


class Endpoint:
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return "<Endpoint {}>".format(self.x)

    __str__ = __repr__


class Middleware:
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return "<Middleware {}>".format(self.x)

    __str__ = __repr__


@pytest.fixture
def endpoint_factory():
    return Endpoint


@pytest.fixture
def middleware_factory():
    return Middleware



from string import punctuation


cdef:
    str GLOB = '*'

    inline int _get_position(int i, int n):
        if i == -1:
            return n
        return i

    inline int _findinstr(str source, str word, int start):
        return source.find(word, start)

    inline void _append_no_conflict_handlers_if_any(RadixTreeNode root, str method, list nc_handlers):
        if root.no_conflict_methods.__contains__(method):
            nc_handlers.extend(root.no_conflict_methods[method])


cdef class RadixTreeNode:
    cdef:
        public str path
        public dict methods
        public dict no_conflict_methods
        public list children
        public str indices
        public int indices_len
        public int path_len
        public dict optimized_index

    def __cinit__(self, str path=None, object handler=None, list methods=None):
        if path is None:
            self.path_len = 0
        else:
            self.path_len = len(path)
        self.path = path
        self.methods = dict()
        self.no_conflict_methods = dict()
        self.children = list()
        self.indices = ""
        self.indices_len = 0
        self.optimized_index = dict()

        self.add_methods(methods, handler)

    def __repr__(self):
        return (
            "<RadixTreeNode path: {}, methods: {}, indices: \"{}\", children: "
            "{}>".format(
                "\"{}\"".format(self.path) if self.path is not None else "None",
                self.methods,
                self.indices,
                self.children
            )
        )

    def add_methods(self, methods, handler, no_conflict=False):
        if not methods:
            return

        if no_conflict:
            for method in methods:
                if method in self.no_conflict_methods and handler in self.no_conflict_methods[method]:
                    continue

                if method in self.no_conflict_methods:
                    self.no_conflict_methods[method].append(handler)
                else:
                    self.no_conflict_methods[method] = [handler]
        else:
            for method in methods:
                if method in self.methods and self.methods[method] != handler:
                    raise KeyError(
                        "{} conflicts with existing handler "
                        "{}".format(handler, self.methods[method])
                    )

                self.methods[method] = handler

    cpdef RadixTreeNode insert_child(self, str index, RadixTreeNode child):
        return self._c_insert_child(index, child)

    cdef RadixTreeNode _c_insert_child(self, str index, RadixTreeNode child):
        cdef int pos = self.get_index_position(index)
        self.indices = self.indices[:pos] + index + self.indices[pos:]
        self.indices_len = len(self.indices)
        self.children.insert(pos, child)
        return child

    cdef int get_index_position(self, str index):
        return _get_position(self.indices.find(index), 0)

    cdef RadixTreeNode get_child(self, str index):
        cdef int i = self.get_index_position(index)
        if i < self.indices_len and self.indices[i] == index:
            return self.children[i]

    cdef optimize(self):
        cdef:
            str index
            RadixTreeNode child

        self.optimized_index.clear()  # if not already clean

        for index in self.indices:
            self.optimized_index[index] = self.get_child(index)

        for child in self.children:
            child.optimize()

    cdef RadixTreeNode get_child_optimized(self, str index):
        return self.optimized_index.get(index)


cdef class RadixTree:
    cdef:
        public RadixTreeNode root
        str _VARIABLE
        str _SEPARATOR

    def __cinit__(self, str variable=None, str separator=None):
        self.root = RadixTreeNode()
        self.VARIABLE = variable or ':'
        self.SEPARATOR = separator or '/'

    def __repr__(self):
        return repr(self.root)

    @property
    def config(self):
        return {
            'variable': self.VARIABLE,
            'separator': self.SEPARATOR
        }

    @property
    def SEPARATOR(self):
        return self._SEPARATOR

    @SEPARATOR.setter
    def SEPARATOR(self, str value):
        if self.root.children:
            raise ValueError("You can't change the separator character after routes have been inserted")
        if value is not None and value in punctuation and len(value) == 1:
            if value == self._VARIABLE:
                raise ValueError("The separator character must not equal the variable character")
            self._SEPARATOR = value
        else:
            raise ValueError("The separator character must be of length one and a valid punctuation")

    @property
    def VARIABLE(self):
        return self._VARIABLE

    @VARIABLE.setter
    def VARIABLE(self, str value):
        if self.root.children:
            raise ValueError("You can't change the variable character after routes have been inserted")
        if value is not None and value in punctuation and len(value) == 1:
            if value == self._SEPARATOR:
                raise ValueError("The variable character must not equal the separator character")
            self._VARIABLE = value
        else:
            raise ValueError("The variable character must be of length one and a valid punctuation")

    def insert(self, str path, object handler, list methods, bint no_conflict=False):
        if path is None or len(path.strip()) == 0 or path.strip()[0] != self._SEPARATOR:
            raise ValueError("path cannot be None, empty or invalid")

        i, node = self._c_insert(path, handler, methods, no_conflict)

        if node is not None:
            conflict = [path[:i] + p for p in self.traverse(node)]
            raise ValueError('"{}" conflicts with {}'.format(path, conflict))

        self.root.optimize()

    cdef list traverse(self, RadixTreeNode root):
        cdef:
            list r = []
            RadixTreeNode child

        if len(root.indices) != 0:
            for i, c in enumerate(root.indices):
                child = root.children[i]
                path = "{}{}".format(c if c in [self._VARIABLE, GLOB] else "", child.path)

                if child.methods and child.indices:
                    r.append(path)

                r.extend([path + p for p in self.traverse(child) or [""]])

        return r

    cdef tuple _c_insert(self, str path, object handler, list methods, bint no_conflict):
        cdef:
            int i = 0
            int n = len(path)
            int code = 0
            int j
            int p
            int m
            RadixTreeNode root
            RadixTreeNode child

        root = self.root

        while i < n:
            if not no_conflict:
                if len(root.indices) != 0 and (root.indices[0] == GLOB or
                    path[i] == GLOB and len(root.indices) != 0 or
                    path[i] != self._VARIABLE and root.indices[0] == self._VARIABLE or
                    path[i] == self._VARIABLE and root.indices[0] != self._VARIABLE or
                    path[i] == self._VARIABLE and root.indices[0] == self._VARIABLE and
                    path[i + 1: _get_position(_findinstr(path, self._SEPARATOR, i), n)] != root.children[0].path):

                    return i, root

            child = root.get_child(path[i])

            if child is None:
                p = _get_position(path.find(self._VARIABLE, i), n)
                if p == n:
                    p = _get_position(path.find(GLOB, i), n)
                    if p == n:
                        root.insert_child(path[i], RadixTreeNode(path[i:], handler, methods))
                        return code, None

                    root = root.insert_child(path[i], RadixTreeNode(path[i:p]))
                    root.insert_child(GLOB, RadixTreeNode(path[p + 1:], handler, methods))
                    return code, None

                root = root.insert_child(path[i], RadixTreeNode(path[i:p]))
                i = _get_position(path.find(self._SEPARATOR, p), n)
                root = root.insert_child(self._VARIABLE, RadixTreeNode(path[p + 1: i]))

                if i == n:
                    root.add_methods(methods, handler, no_conflict)
            else:
                root = child
                if path[i] == self._VARIABLE:
                    i += len(root.path) + 1
                    if i == n:
                        root.add_methods(methods, handler, no_conflict)
                else:
                    j = 0
                    m = len(root.path)

                    while i < n and j < m and path[i] == root.path[j]:
                        i += 1
                        j += 1

                    if j < m:
                        child = RadixTreeNode(root.path[j:])
                        child.methods = root.methods
                        child.no_conflict_methods = root.no_conflict_methods
                        child.children = root.children
                        child.indices = root.indices
                        child.indices_len = len(child.indices)

                        root.path = root.path[:j]
                        root.path_len = len(root.path)
                        root.methods = {}
                        root.no_conflict_methods = {}
                        root.children = [child]
                        root.indices = child.path[0]
                        root.indices_len = len(root.indices)

                    if i == n:
                        root.add_methods(methods, handler, no_conflict)

        return code, None

    cpdef tuple get(self, str path, str method):
        cdef dict params = {}

        handler, nc_handlers = self._c_get(path, method, params)
        if handler is None:
            return None, [], {}
        return handler, nc_handlers, params

    cdef tuple _c_get(self, str path, str method, dict params):
        cdef:
            int i = 0
            int n = len(path)
            int pos
            object handler = None
            list nc_handlers = []
            RadixTreeNode root

        root = self.root

        while i < n:
            if root.indices_len == 0:
                return None, None

            _append_no_conflict_handlers_if_any(root, method, nc_handlers)

            if root.indices[0] == self._VARIABLE:
                root = root.children[0]
                pos = _get_position(_findinstr(path, self._SEPARATOR, i), n)
                params[root.path] = path[i:pos]
                i = pos

            elif root.indices[0] == GLOB:
                root = root.children[0]
                params[root.path] = path[i:]
                break
            else:
                root = root.get_child_optimized(path[i])

                if root is None:
                    return None, None

                pos = i + root.path_len

                if path[i:pos] != root.path:
                    return None, None

                i = pos

        handler = root.methods.get(method)

        if handler is None:
            return None, None

        _append_no_conflict_handlers_if_any(root, method, nc_handlers)

        return root.methods.get(method), nc_handlers



import pytest

from xrtr import RadixTree


def test_tree():
    tree = RadixTree()
    assert tree is not None
    assert tree.root is not None
    assert len(tree.root.children) == 0
    assert tree.root.path is None
    assert tree.root.indices == ""
    assert tree.root.methods == {}
    assert tree.root.no_conflict_methods == {}

    r = tree.get("/foo", "BAR")
    assert r[0] is None
    assert r[1] == []
    assert r[2] == {}


def test_tree_repr(endpoint_factory):
    endpoint_1 = endpoint_factory(1)
    tree = RadixTree()
    assert repr(tree) == repr(tree.root)
    assert (
        repr(tree)
        == '<RadixTreeNode path: None, methods: {}, indices: "", children: []>'
    )
    assert (
        repr(tree.root)
        == '<RadixTreeNode path: None, methods: {}, indices: "", children: []>'
    )
    tree.insert("/foo", endpoint_1, ["BAR"])
    assert (
        repr(tree.root.children[0])
        == '<RadixTreeNode path: "/foo", methods: {\'BAR\': <Endpoint 1>}, indices: "", children: []>'
    )


def test_tree_validations(endpoint_factory):
    endpoint_1 = endpoint_factory(1)
    tree = RadixTree()

    with pytest.raises(ValueError):
        tree.insert(None, endpoint_1, ["BAR"])

    with pytest.raises(ValueError):
        tree.insert("", endpoint_1, ["BAR"])

    with pytest.raises(ValueError):
        tree.insert("     ", endpoint_1, ["BAR"])

    with pytest.raises(ValueError):
        tree.insert("foo", endpoint_1, ["BAR"])

    with pytest.raises(ValueError):
        tree.insert("   foo   ", endpoint_1, ["BAR"])

    with pytest.raises(ValueError):
        tree.insert("f/oo/", endpoint_1, ["BAR"])


def test_tree_config_1(endpoint_factory):
    endpoint_1 = endpoint_factory(1)
    tree = RadixTree()
    tree.VARIABLE = "$"
    tree.SEPARATOR = "."

    tree.insert(".foo", endpoint_1, ["BAR"])
    tree.insert(".foo.$bar", endpoint_1, ["BAR"])
    tree.insert(".bar", endpoint_1, ["BAR"])
    tree.insert(".bar.*rest", endpoint_1, ["BAR"])

    r = tree.get(".foo.hello", "BAR")
    assert r[0] == endpoint_1
    assert r[1] == []
    assert r[2] == {"bar": "hello"}

    r = tree.get(".bar.avra.cadavra.alacazam", "BAR")
    assert r[0] == endpoint_1
    assert r[1] == []
    assert r[2] == {"rest": "avra.cadavra.alacazam"}

    assert tree.config == {"variable": "$", "separator": "."}


def test_tree_config_2(endpoint_factory):
    endpoint_1 = endpoint_factory(1)
    tree = RadixTree(variable="$", separator=".")

    tree.insert(".foo", endpoint_1, ["BAR"])
    tree.insert(".foo.$bar", endpoint_1, ["BAR"])
    tree.insert(".bar", endpoint_1, ["BAR"])
    tree.insert(".bar.*rest", endpoint_1, ["BAR"])

    r = tree.get(".foo.hello", "BAR")
    assert r[0] == endpoint_1
    assert r[1] == []
    assert r[2] == {"bar": "hello"}

    r = tree.get(".bar.avra.cadavra.alacazam", "BAR")
    assert r[0] == endpoint_1
    assert r[1] == []
    assert r[2] == {"rest": "avra.cadavra.alacazam"}

    assert tree.config == {"variable": "$", "separator": "."}


def test_tree_config_error():
    tree = RadixTree()

    with pytest.raises(ValueError):
        tree.VARIABLE = "/"

    with pytest.raises(ValueError):
        tree.VARIABLE = "a"

    with pytest.raises(ValueError):
        tree.VARIABLE = "$$"

    with pytest.raises(ValueError):
        tree.VARIABLE = ""

    with pytest.raises(ValueError):
        tree.VARIABLE = None

    with pytest.raises(ValueError):
        tree.SEPARATOR = ":"

    with pytest.raises(ValueError):
        tree.SEPARATOR = "a"

    with pytest.raises(ValueError):
        tree.SEPARATOR = ".."

    with pytest.raises(ValueError):
        tree.SEPARATOR = ""

    with pytest.raises(ValueError):
        tree.SEPARATOR = None

    with pytest.raises(ValueError):
        RadixTree(variable="/")

    with pytest.raises(ValueError):
        RadixTree(separator=":")

    with pytest.raises(ValueError):
        RadixTree(variable=":", separator=":")

    with pytest.raises(ValueError):
        RadixTree(variable="a", separator="b")

    with pytest.raises(ValueError):
        RadixTree(separator="b")

    with pytest.raises(ValueError):
        RadixTree(variable="a")

    with pytest.raises(ValueError):
        RadixTree(variable="::", separator="//")

    with pytest.raises(ValueError):
        RadixTree(variable="::")

    with pytest.raises(ValueError):
        RadixTree(separator="//")

    # ----------------------------------------------------------------------- #
    # all tests bellow will assume defaults
    # ----------------------------------------------------------------------- #

    RadixTree(variable="", separator="")
    RadixTree(separator="")
    RadixTree(separator="")
    RadixTree(variable=None, separator=None)
    RadixTree(variable=None)
    RadixTree(separator=None)

    tree.insert("/", object(), ["BAR"])

    with pytest.raises(ValueError):
        tree.SEPARATOR = "."

    with pytest.raises(ValueError):
        tree.VARIABLE = "$"


def test_tree_single_endpoint(endpoint_factory):
    endpoint_1 = endpoint_factory(1)
    tree = RadixTree()

    tree.insert("/foo/:bar", endpoint_1, ["BAR"])

    r = tree.get("/foo/hello", "BAR")
    assert r[0] == endpoint_1
    assert r[1] == []
    assert r[2] == {"bar": "hello"}

    r = tree.get("/foooo", "BAR")
    assert r[0] is None
    assert r[1] == []
    assert r[2] == {}


def test_tree_root_endpoint(endpoint_factory):
    endpoint_1 = endpoint_factory(1)
    tree = RadixTree()

    tree.insert("/", endpoint_1, ["BAR"])

    r = tree.get("/", "BAR")
    assert r[0] == endpoint_1
    assert r[1] == []
    assert r[2] == {}


def test_tree_full(endpoint_factory, middleware_factory):
    endpoint_1 = endpoint_factory(1)
    middleware_1 = middleware_factory(1)
    middleware_2 = middleware_factory(2)
    middleware_3 = middleware_factory(3)
    middleware_4 = middleware_factory(4)

    tree = RadixTree()

    tree.insert("/foo", endpoint_1, ["FOO"])
    tree.insert("/foo", middleware_1, ["FOO", "BAR"], no_conflict=True)
    tree.insert("/foo", middleware_1, ["BAR"], no_conflict=True)
    tree.insert("/foo/:name", endpoint_1, ["FOO", "BAR"])
    tree.insert(
        "/foo/:name/:x", middleware_2, ["FOO", "BAR"], no_conflict=True
    )
    tree.insert("/foo/:name/:x/:y", endpoint_1, ["FOO", "BAR"])
    tree.insert("/static/*path", endpoint_1, ["FOO"])

    with pytest.raises(ValueError):
        tree.insert("/foo/:bar", endpoint_1, ["FOO"])

    with pytest.raises(KeyError):
        tree.insert("/foo", middleware_2, ["FOO"])

    r = tree.get("/foo/hello", "FOO")
    assert r[0] == endpoint_1
    assert r[1] == [middleware_1]
    assert r[2] == {"name": "hello"}

    r = tree.get("/foo/hello/a/b", "BAR")
    assert r[0] == endpoint_1
    assert r[1] == [middleware_1, middleware_2]
    assert r[2] == {"name": "hello", "x": "a", "y": "b"}

    r = tree.get("/static/path/to/my/file.py", "FOO")
    assert r[0] == endpoint_1
    assert r[1] == []
    assert r[2] == {"path": "path/to/my/file.py"}

    tree.insert(
        "/foo/:name/:x", middleware_3, ["FOO", "BAR"], no_conflict=True
    )

    tree.insert("/foo/:name/:x/:y", middleware_4, ["FOO"], no_conflict=True)

    r = tree.get("/foo/hello/a/b", "FOO")
    assert r[0] == endpoint_1
    assert r[1] == [middleware_1, middleware_2, middleware_3, middleware_4]
    assert r[2] == {"name": "hello", "x": "a", "y": "b"}

    r = tree.get("/foo/hello/aa", "FOO")
    assert r[0] is None
    assert r[1] == []
    assert r[2] == {}



import io
import os
import re
import sys

from setuptools import Extension
from setuptools import find_packages
from setuptools import setup

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None


_dev_build = "--inplace" in sys.argv
_ci_build = "egg_info" in sys.argv or "install" in sys.argv
_define_macros = []
_extra_cargs = []


if int(os.getenv("FOR_PRODUCTION", "0")) == 1:
    _dev_build = False
    _ci_build = False

    # _extra_cargs.append("-O2")
    # _extra_cargs.append("-ffloat-store")
    # _extra_cargs.append("-fgcse")
    # _extra_cargs.append("-floop-parallelize-all")
    # _extra_cargs.append("-fivopts")
    # _extra_cargs.append("-freorder-blocks-and-partition")
    # _extra_cargs.append("-foptimize-sibling-calls")
    # _extra_cargs.append("-freorder-functions")
    # _extra_cargs.append("-ftree-vectorize")

    _extra_cargs.append("-O3")
    # _extra_cargs.append("-foptimize-strlen")
    _extra_cargs.append("-ffloat-store")
    _extra_cargs.append("-fpartial-inlining")
    # _extra_cargs.append("-fgcse")
    # _extra_cargs.append("-fivopts")
    _extra_cargs.append("-freorder-blocks-and-partition")
    _extra_cargs.append("-foptimize-sibling-calls")
    _extra_cargs.append("-freorder-functions")
    _extra_cargs.append("-flto")

else:
    _extra_cargs.append("-Og")
    _define_macros += [
        ("CYTHON_PROFILE", 1),
        ("CYTHON_TRACE", 1),
        ("CYTHON_TRACE_NOGIL", 1),
    ]


def read(*names, **kwargs):
    return io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8"),
    ).read()


ext_modules = [
    Extension(
        "xrtr",
        sources=["src/xrtr.{}".format("c" if _ci_build else "pyx")],
        extra_compile_args=_extra_cargs,
        define_macros=_define_macros,
    )
]

cy_modules = None

if cythonize is not None:
    cy_modules = cythonize(
        ext_modules,
        compiler_directives={
            "optimize.use_switch": False,
            "optimize.unpack_method_calls": False,
            "language_level": 3,
            "c_string_type": "bytes",
            "c_string_encoding": "ascii",
            "boundscheck": False,
            "wraparound": False,
            "embedsignature": _dev_build,
            "linetrace": _dev_build,
            "profile": _dev_build,
            "binding": _dev_build,
        },
    )

setup(
    name="xrtr",
    version="0.1.1",
    description="A Radix Tree based router for HTTP and other routing needs "
    "with support for middlewares and endpoints with a Cython "
    "boost",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", read("README.rst")
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    author="Richard Kuesters",
    author_email="rkuesters@gmail.com",
    url="https://github.com/vltr/xrtr",
    setup_requires=["Cython"],
    install_requires=["setuptools>=19.0"],
    ext_modules=cy_modules,
    package_data={
        "xrtr": [
            "src/xrtr.pyx",
            "src/xrtr.c",
            "src/__init__.py",
            "src/__init__.pxd",
        ]
    },
    packages=find_packages("src"),
    package_dir={"": "src"},
    zip_safe=False,
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        # 'Programming Language :: Python :: Implementation :: PyPy',
        "Topic :: Utilities",
    ],
    keywords=[
        "router",
        "radix",
        "tree",
        "cython",
        "trie",
        "middleware",
        "endpoint",
    ],
)
